import getpass
import numpy as np
import sklearn

from sklearn.metrics import accuracy_score
from sklearn.preprocessing import StandardScaler
from sklearn.cross_validation import StratifiedKFold

from calculate_five_fold_curves import cadd_scores
from calculate_five_fold_curves import polyphen2_scores
from calculate_five_fold_curves import sift_scores
from calculate_five_fold_curves import dbnsfp_svm_scores
from calculate_five_fold_curves import dbnsfp_lr_scores

NUM_FOLDS = 5   # Number of folds for cross validation
HGMD_THRESHOLD = .98
#HGMD_FEATURES = "/cluster/u/{}/pathogenicity_classification/data/extract_hgmd.features.txt"
#KGP_FEATURES = "/cluster/u/{}/pathogenicity_classification/data/extract_kgp.features.txt"
HGMD_FEATURES = "/cluster/u/{}/pathogenicity_classification/data/hgmd.train.txt"
KGP_FEATURES = "/cluster/u/{}/pathogenicity_classification/data/exac.train.txt"
HGMD_FEATURES = HGMD_FEATURES.format(getpass.getuser())
KGP_FEATURES = KGP_FEATURES.format(getpass.getuser())

def get_data(hgmd_fname=HGMD_FEATURES, kgp_fname=KGP_FEATURES):
    hgmd_data = np.genfromtxt(hgmd_fname, delimiter="\t", dtype=np.dtype(str), skip_header=1)
    kgp_data = np.genfromtxt(kgp_fname, delimiter="\t", dtype=np.dtype(str), skip_header=1)
    return hgmd_data, kgp_data

def get_dataset(with_metrics, hgmd_fname=HGMD_FEATURES, kgp_fname=KGP_FEATURES):
    hgmd, kgp = get_data(hgmd_fname=hgmd_fname, kgp_fname=kgp_fname)
    hgmd = hgmd[:,4:-5].astype(float)
    kgp = kgp[:,4:-5].astype(float)
    X = np.vstack((hgmd, kgp))
    Y = np.vstack((np.ones((len(hgmd),1)), np.zeros((len(kgp),1))))

    if with_metrics:
        cadd, _ = cadd_scores()
        sift, _ = sift_scores()
        polyphen2, _ = polyphen2_scores()
        svm, _ = dbnsfp_svm_scores()
        lr, _ = dbnsfp_lr_scores()
        X = np.hstack((X, sift, cadd, polyphen2, svm, lr))

    return X, Y

def get_feature_names():
    with open(HGMD_FEATURES) as f:
        header = f.readline().strip('\n')
        return header.split('\t')

def write_data(fname, data):
    np.savetxt(fname, data, delimiter='\t', fmt='%s')

def write_probs(clf_name, hgmd_data, kgp_data):
    start, end = HGMD_FEATURES.rsplit('.', 1)
    fname = '.'.join([start, '{}_probs'.format(clf_name), end])
    write_data(fname, hgmd_data)

    start, end = KGP_FEATURES.rsplit('.', 1)
    fname = '.'.join([start, '{}_probs'.format(clf_name), end])
    write_data(fname, kgp_data)

def calculate_threshold(clf, X, Y):
    """
    Calculate a decision boundary in order to ensure that we have HGMD_THRESHOLD
    accuracy on the hgmd_test set.

    clf: A trained classifier which has the 'predict_proba' function.
    """
    hgmd_test = X[np.where(Y == 1)[0]]
    hgmd_test_labels = Y[np.where(Y == 1)[0]]

    kgp_test = X[np.where(Y == 0)[0]]
    baylor_test_labels = Y[np.where(Y == 0)[0]]

    thresholds = np.arange(0, 1.01, .01)
    best_threshold = 0.0
    best_hgmd_acc = 0.0
    best_baylor_acc = 0.0

    hgmd_probs = clf.predict_proba(hgmd_test)
    baylor_probs = clf.predict_proba(kgp_test)
    
    for threshold in thresholds:
        hgmd_preds = (hgmd_probs > threshold).astype('float')[:,1]
        hgmd_acc = accuracy_score(hgmd_test_labels.flatten(), hgmd_preds)

        baylor_preds = (baylor_probs > threshold).astype('float')[:,1]
        baylor_acc = accuracy_score(baylor_test_labels.flatten(), baylor_preds)

        if baylor_acc > best_baylor_acc and hgmd_acc >= HGMD_THRESHOLD:
            best_threshold = threshold
            best_hgmd_acc = hgmd_acc
            best_baylor_acc = baylor_acc

    return (best_threshold, best_hgmd_acc, best_baylor_acc)

def evaluate(clf, processor, with_metrics=False, save_probs=False, return_clfs=False):
    """
    Evaluate the clf with 5-fold cross validation.

    clf: a classifier which has 'fit' and 'predict_proba' functions
    processor: a function which preprocesses the data before fitting and testing
    """
    X, Y = get_dataset(with_metrics)

    hgmd_data_with_probs = []
    kgp_data_with_probs = []
    results = []
    trained_clfs = []

    skf = StratifiedKFold(Y.flatten(), n_folds=NUM_FOLDS, shuffle=True)
    for train_idxs, test_idxs in skf:
        X_train, X_test = X[train_idxs], X[test_idxs]
        Y_train, Y_test = Y[train_idxs], Y[test_idxs]

        X_train, X_test, _ = processor(X_train, X_test, np.array([X_train[0]]))

        # Get a new copy of the clf that hasn't been fitted yet
        clf = sklearn.base.clone(clf)

        # Fit the classifier and test on the held out fold
        clf.fit(X_train, Y_train.flatten())
        results.append(calculate_threshold(clf, X_test, Y_test))
        print(results[-1])

        # Add the classifier probabilities for the held out set to the data matrix
        if save_probs:
            hgmd_probs = clf.predict_proba(X_test[np.where(Y_test == 1)[0]])
            hgmd_data_with_probs.append(np.hstack((X_test[np.where(Y_test == 1)[0]], hgmd_probs[:,1:])))
            kgp_probs = clf.predict_proba(X_test[np.where(Y_test == 0)[0]])
            kgp_data_with_probs.append(np.hstack((X_test[np.where(Y == 0)[0]], kgp_probs[:,1:])))

        # Keep this clf to return later
        if return_clfs:
            trained_clfs.append(clf)

    if save_probs:
        hgmd_data_with_probs = np.vstack(hgmd_data_with_probs)
        kgp_data_with_probs = np.vstack(kgp_data_with_probs)
        write_probs(clf.__class__.__name__, hgmd_data_with_probs, kgp_data_with_probs)

    means = np.mean(results, axis=0)
    stds = np.std(results, axis=0)
    print('threshold:\t{:.5f} +/- {:.5f}'.format(means[0], stds[0]))
    print('hgmd_acc:\t{:.5f} +/- {:.5f}'.format(means[1], stds[1]))
    print('baylor_acc:\t{:.5f} +/- {:.5f}'.format(means[2], stds[2]))

    if return_clfs:
        return trained_clfs
